# -*- coding: utf-8 -*-

import logging

import flask
from flask import Blueprint, render_template
from flask_restplus import Api, Resource, apidoc, inputs
from pyris import address
from pyris.api import extract
from pyris.api.insee import api as insee_api

Logger = logging.getLogger(__name__)

service = Blueprint(
    'api',
    __name__)


@service.route('/')
def index():
    return render_template("index.html")

@service.route('/map')
def map():
    return render_template("map.html")

@service.route('/iris/<code>')
def iris(code):
    data = extract.get_complete_iris(code)
    if not data:
        return 404, "IRIS code '{}' not found".format(code)
    print(data)
    return render_template("iris.html",
                           iris=data['complete_code'],
                           name=data['name'],
                           citycode=data['citycode'],
                           city=data['city'],
                           iris_type=data['type'])


api = Api(service,
          title='INSEE/IRIS Geolocalizer',
          ui=False,
          prefix='/api',
          version='0.1',
          description="Retrieve some data related to the IRIS codes. Look for an IRIS from an address.")
api.add_namespace(insee_api)

geojson_parser = api.parser()
geojson_parser.add_argument("geojson", type=inputs.boolean, default=False, location='args',
                            help='GeoJSON')

iris_code_parser = geojson_parser.copy()
iris_code_parser.add_argument("limit", required=False, default=10, dest='limit',
                              location='args', help='Limit')

address_parser = geojson_parser.copy()
address_parser.add_argument("q", required=True, dest='q', location='args',
                            help='Query')
address_parser.add_argument("postcode", required=False, dest='postcode', location='args',
                            help='postcode (code Postal)')
address_parser.add_argument("citycode", required=False, dest='citycode', location='args',
                            help='citycode (code INSEE)')
address_parser.add_argument("lat", required=False, type=float, dest='lat', location='args',
                            help='latitude')
address_parser.add_argument("lon", required=False, type=float, dest='lon', location='args',
                            help='longitude')
address_parser.add_argument("limit", required=False, type=int, dest='limit', location='args',
                            help='limite (nombre d’élements retournés)')


coords_parser = geojson_parser.copy()
coords_parser.add_argument("lat", required=True, type=float, dest='lat', location='args',
                            help='Latitude')
coords_parser.add_argument("lon", required=True, type=float, dest='lon', location='args',
                            help='Longitude')


@service.route('/doc/')
def swagger_ui():
    return apidoc.ui_for(api)


@api.route("/iris/<string:code>")
class IrisCode(Resource):
    @api.doc(parser=iris_code_parser,
             description="get data for a specific IRIS (four digits)")
    def get(self, code):
        args = iris_code_parser.parse_args()
        limit = args['limit']
        Logger.info("look for IRIS '%s'", code)
        Logger.info("with limit %s", limit)
        iris = extract.get_iris_field(code, limit, args['geojson'])
        if not iris:
            api.abort(404, "IRIS code '{}' not found".format(code))
        return iris


@api.route('/city/code/<string:code>')
class IrisListFromCityCode(Resource):
    @api.doc(parser=api.parser(),
             description="Look for all the iris codes in the city matching a city code")
    def get(self, code):
        Logger.info("looking for IRIS in the city %s", code)
        iris=extract.get_iris_list_by_city_code(code)
        if not iris:
            api.abort(404, "City code '{}' not found".format(code))
        return iris


@api.route("/city/search/")
class IrisListFromQuery(Resource):
    @api.doc(parser=address_parser,
             description="Get the list of all the iris codes in the city matching the query")
    def get(self):
        args = address_parser.parse_args()
        q, postcode, citycode, lat, lon, limit = args['q'], args['postcode'], args['citycode'], args['lat'], args['lon'], args['limit']
        Logger.info("Look for a list of IRIS in a city for q '%s', postcode '%s', citycode '%s', lat '%s' , lon '%s', limit '%s'",
                    q, postcode, citycode, lat, lon, limit)
        coord = address.coordinate(q, postcode, citycode, lat, lon, limit)
        if coord["address"] is None:
            # Requests sent to '/api/search' that match nothing do not return a 404 error like other requests.
            # Is this an intentional choice?
            api.abort(404, "No city found matching that query")
        # I'm pretty sure the preferred order is usually latitude first, then longitude
        Logger.info("Looking for iris at coordinates %s, %s", coord["lat"], ["lon"])
        iris = extract.iris_from_coordinate(coord["lon"], coord["lat"])
        iris_list = extract.get_iris_list_by_city_code(iris["citycode"])
        return {"city_name": iris["city"], "city_code": iris["citycode"], "iris_list": iris_list}


@api.route("/coords")
class IrisFromCoordinates(Resource):
    @api.doc(parser=coords_parser,
             description="get IRIS from a lat/lon coordinates")
    def get(self):
        args = coords_parser.parse_args()
        lat, lon = args['lat'], args['lon']
        Logger.info("Coordinates '%s'", (lon, lat))
        Logger.info("geojson? %s", args['geojson'])
        res = extract.iris_from_coordinate(lon, lat, args['geojson'])
        if not res:
            api.abort(404, "IRIS not found from coordinates ({}, {})".format(lat, lon))
        return res


@api.route("/compiris/<string:code>")
class CompleteIrisCode(Resource):
    @api.doc(parser=geojson_parser,
             description=("Get data for a specific complete IRIS code (9 digits)."
                          " INSEE City code + IRIS code"))
    def get(self, code):
        Logger.info("look for IRIS '%s'", code)
        args = geojson_parser.parse_args()
        iris = extract.get_complete_iris(code, args['geojson'])
        if not iris:
            api.abort(404, "Complete IRIS code '{}' not found".format(code))
        return iris


@api.route("/search/")
class IrisFromAddress(Resource):
    @api.doc(parser=address_parser,
             description="Look for an IRIS for a specific address. This will use the `search` endpoint of API Adresse of geo.api.gouv.fr service (see documentation here geo.api.gouv.fr/adresse) to retrieve coordinates and then will retrieve iris code from those coordinates")
    def get(self):
        args = address_parser.parse_args()
        q, postcode, citycode, lat, lon, limit = args['q'], args['postcode'], args['citycode'], args['lat'], args['lon'], args['limit']
        Logger.info("Look for IRIS for q '%s', postcode '%s', citycode '%s', lat '%s' , lon '%s', limit '%s'", q, postcode, citycode, lat, lon, limit)
        coord = address.coordinate(q, postcode, citycode, lat, lon, limit)
        Logger.info("Get coordinate (%s, %s)", coord["lon"], coord["lat"])
        Logger.info("For address '%s'", coord["address"])
        if coord['address'] is None:
            return []
        res = extract.iris_from_coordinate(coord['lon'], coord['lat'], args['geojson'])
        res.update(coord)
        return res



# -*- coding: utf-8 -*-

import io
import logging
import os
from datetime import date, datetime

from yaml import load as yload

import flask
from flask import Flask
from flask.json import JSONEncoder
from flask_restplus import Api as rpapi
from pyris.api.app import service

# app version
__version__ = '1.0'

ISO_FORMAT = "%Y-%m-%d"
Logger = logging.getLogger(__name__)

LOG_LEVELS = {
    'debug': logging.DEBUG,
    'info': logging.INFO,
    'warning': logging.WARNING,
    'error': logging.ERROR,
    'critical': logging.CRITICAL
}

console = logging.StreamHandler()
log_formatter = logging.Formatter("%(asctime)s :: %(levelname)s :: %(module)s :: %(funcName)s :: %(message)s")
console.setFormatter(log_formatter)
Logger.addHandler(console)


class CustomJSONEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (date, datetime)):
            return obj.strftime(ISO_FORMAT)
        return JSONEncoder.default(self, obj)


def set_level(level='info'):
    """
    Set log level
    """
    Logger.setLevel(LOG_LEVELS.get(level))


def register_extensions(app):
    """
    Register extensions into flask application
    """
    app.register_blueprint(service)


def load_yaml_config(filename):
    """
    Open Yaml file, load content for flask config and returns it as a python dict
    """
    content = io.open(filename, 'r').read()
    return yload(content).get('flask', {})


@property
def specs_url(self):
    """Fixes issue where swagger-ui makes a call to swagger.json over HTTP.
       This can ONLY be used on servers that actually use HTTPS.  On servers that use HTTP,
       this code should not be used at all.
    """
    return flask.url_for(self.endpoint('specs'), _external=True, _scheme='https')


def create_app(env='Defaults'):
    """
    Creates application.
    :returns: flask application instance
    """
    app = Flask(__name__)
    cfgfile = os.environ.get('PYRIS_APP_SETTINGS')
    Logger.info(cfgfile)
    if cfgfile:
        app.config.update(load_yaml_config(cfgfile))

    if 'LOG_LEVEL' not in app.config:
        set_level('debug')
    else:
        set_level(app.config['LOG_LEVEL'])
    if app.config['FORCE_SWAGGER_JSON_HTTPS']:
        rpapi.specs_url = specs_url
    Logger.info("Run the app")
    register_extensions(app)
    app.json_encoder = CustomJSONEncoder
    app.config.SWAGGER_UI_DOC_EXPANSION = 'list'

    return app



# coding: utf-8

"""French address geolocalizer API
"""

from requests.exceptions import ConnectionError

import slumber
from pyris.config import GEOCODER_URL as URL

URL_CALLBACK = "http://api-adresse.data.gouv.fr"

api = slumber.API(URL)
api_callback = slumber.API(URL_CALLBACK)


def _search(q, postcode, citycode, lat, lon, limit):
    """Allow to callback to the national geocoder service if the local one fails
    """
    try:
        resp = api.search.get(q=q, postcode=postcode, citycode=citycode, lat=lat,
                              lon=lon, limit=limit)
        if not resp["features"]:
            resp = api_callback.search.get(q=q, postcode=postcode, citycode=citycode,
                                           lat=lat, lon=lon, limit=limit)
    except ConnectionError:
        resp = api_callback.search.get(q=q, postcode=postcode, citycode=citycode,
                                       lat=lat, lon=lon, limit=limit)
    return resp


def coordinate(q, postcode, citycode, lat, lon, limit):
    """Get lon/lat coordinate from an address

    q: str
    postcode: str
    citycode: str

    Return (longitude, latitde)
    """
    resources = _search(q=q, postcode=postcode, citycode=citycode, lat=lat, lon=lon, limit=limit)
    if len(resources['features']) == 0:
        return {"lon": None,
                "lat": None,
                'address': None}
    lon, lat = lonlat(resources["features"][0])
    return {"lon": lon,
            "lat": lat,
            "address": retrieve_address(resources["features"][0])}


def retrieve_address(feature):
    """Retrieve the address from a feature
    """
    return feature["properties"]["label"]


def lonlat(feature):
    """Longitude and latitude from a GeoJSON feature
    """
    return feature["geometry"]["coordinates"]


if __name__ == '__main__':
    q = "place des quinconces bordeaux"
    res = coordinate(q)



"""Extract data from the database
"""

import os
import json
import logging

import psycopg2

from pyris.config import DATABASE


_HERE = os.path.abspath(os.path.dirname(__file__))
_QUERY_DIR = os.path.join(_HERE, "queries")
Q_IRIS = "iris.sql"
Q_IRIS_BY_CITY_CODE = "iris_by_city_code.sql"
Q_COMPIRIS = "complete_iris.sql"
Q_COORD = "coordinate.sql"
Q_POPULATION = "iris_population.sql"
Q_POPULATION_AGE = "iris_population_age.sql"
Q_POPULATION_SEX = "iris_population_sex.sql"
Q_LOGEMENT = "iris_logement.sql"
Q_LOGEMENT_ROOM = "iris_logement_room.sql"
Q_LOGEMENT_AREA = "iris_logement_area.sql"
Q_LOGEMENT_YEAR = "iris_logement_year.sql"
Q_EMPLOYMENT = "iris_activite.sql"
Q_EMPLOYMENT_SEX = "iris_activite_sex.sql"
Q_EMPLOYMENT_AGE = "iris_activite_age.sql"
Q_EMPLOYMENT_SECTOR = "iris_activite_secteur.sql"

Logger = logging.getLogger(__name__)


def _load_sql_file(fname):
    """Return the content of the SQL file `fname`

    fname: str

    Return a string
    """
    skip = lambda x: x.strip().startswith('--') or len(x.strip()) == 0
    with open(os.path.join(_QUERY_DIR, fname)) as fobj:
        return "".join(line for line in fobj if not skip(line))


def _query(q, params=None, columns=False):
    """Carry out a SQL query

    Only fetch one result
    """
    Logger.debug("processing query '%s'", q)
    with psycopg2.connect(database=DATABASE["DBNAME"],
                          user=DATABASE['USER'],
                          password=DATABASE.get('PASSWORD'),
                          host=DATABASE['HOST']) as cnx:
        with cnx.cursor() as cu:
            if params is not None:
                cu.execute(q, params)
            else:
                cu.execute(q)
            rset = cu.fetchall()
            if rset and columns:
                names = [x.name for x in cu.description]
                return [{x: v for x, v in zip(names, values)}
                        for values in rset]
            return rset


def _iris_fields(res, geojson=False):
    """Iris field from a SQL query result
    """
    data = {"iris": res[0],
            'city': res[1],
            'citycode': res[2],
            'name': res[3],
            'complete_code': res[4],
            'type': res[5]}
    if geojson:
        return {"type": "Feature",
                "geometry": json.loads(res[6]),
                "properties": data}
    return data


def _split_data(data):
    """Split the keys of the data between 'properties' and 'data'

    data: list (or None)
        Result of a SQL query

    Returns
    -------
    dict
    """
    if data is None:
        return data
    data = data[0]
    properties = ["iris", "city", "citycode", "label"]
    result = {k: data.pop(k) for k in properties}
    result['data'] = data
    return result


def get_iris_field(code, limit=None, geojson=False):
    """Get some data from the IRIS code

    code: str
        IRIS code. Four digits
    limit: int (None)
        number of results
    """
    query_iris = _load_sql_file(Q_IRIS)
    if limit is not None:
        query_iris = query_iris.replace(";", " LIMIT {};".format(limit))
    res = _query(query_iris, (code,))
    Logger.debug("res: %s", res)
    if res:
        data = [_iris_fields(x, geojson) for x in res]
        if geojson:
            return {"type": "FeatureCollection",
                    "features": data}
        return data
    return res


def get_iris_list_by_city_code(code):
    """Get the list of IRIS in a city by its code
    
    code: str
        City code. Five digits.
    """
    query=_load_sql_file(Q_IRIS_BY_CITY_CODE)
    Logger.debug("Query: %s", query)
    res=_query(query, (code,))
    Logger.debug("res: %s", res)
    return [x[0] for x in res] if res else res


def get_complete_iris(code, geojson=False):
    """Get some date from the complete IRIS code

    Complete IRIS code is made up of:
        - INSEE City code (5 digits). Different from postal code
        - IRIS code (4 diits)

    code: str
        Complete IRIS code. Nine digits
    """
    query = _load_sql_file(Q_COMPIRIS)
    Logger.debug("Query '%s'", query)
    res = _query(query, (code,))
    Logger.debug("res: %s", res)
    if res:
        return _iris_fields(res[0], geojson)
    return res


def iris_from_coordinate(lon, lat, geojson=False):
    """Get the IRIS code from a coordinate.
    """
    query_coordinate = _load_sql_file(Q_COORD)
    Logger.debug("Query '%s'", query_coordinate)
    res = _query(query_coordinate, (lon, lat))
    Logger.debug("res: %s", res)
    if res:
        return _iris_fields(res[0], geojson)
    return res


def get_iris_population(code):
    """Get the population for a specific IRIS

    Parameters
    ----------
    code : str
        IRIS code (9 digits)

    Returns
    -------
    list of dicts
    """
    query_population = _load_sql_file(Q_POPULATION)
    Logger.debug("Query '%s'", query_population)
    return _query(query_population, (code,), columns=True)


def get_iris_population_age(code):
    """Get the population distribution by age for a specific IRIS

    Parameters
    ----------
    code : str
        IRIS code (9 digits)

    Returns
    -------
    list of dicts
    """
    query_population = _load_sql_file(Q_POPULATION_AGE)
    data = _query(query_population, (code,), columns=True)
    return _split_data(data)


def get_iris_population_sex(code):
    """Get the population distribution by sex and age for a specific IRIS

    Parameters
    ----------
    code : str
        IRIS code (9 digits)

    Returns
    -------
    dict
    """
    query_population = _load_sql_file(Q_POPULATION_SEX)
    data = _query(query_population, (code,), columns=True)
    return _split_data(data)


def get_iris_logement(code, by=None):
    """Get the housing data for a specific IRIS

    Parameters
    ----------
    code : str
        IRIS code (9 digits)
    by : str (optional)
        Get data by room, area or year

    Returns
    -------
    dict
    """
    if by not in (None, 'room', 'area', 'year'):
        raise ValueError("Value {} for the 'by' parameter is not supported".format(by))
    if by is None:
        query = _load_sql_file(Q_LOGEMENT)
        rset = _query(query, (code,), columns=True)
        return rset[0] if rset else rset
    elif by == 'room':
        query = _load_sql_file(Q_LOGEMENT_ROOM)
    elif by == 'area':
        query = _load_sql_file(Q_LOGEMENT_AREA)
    elif by == 'year':
        query = _load_sql_file(Q_LOGEMENT_YEAR)
    data = _query(query, (code,), columns=True)
    return _split_data(data)


def get_iris_employment(code, by=None):
    """Get the employment data for a specific IRIS

    Parameters
    ----------
    code : str
        IRIS code (9 digits)
    by : str (optional)
        Get data by sex, age

    Returns
    -------
    list of dicts
    """
    if by not in (None, 'sex', 'age', 'sector'):
        raise ValueError("Value {} for the 'by' parameter is not supported".format(by))
    if by is None:
        query = _load_sql_file(Q_EMPLOYMENT)
        rset = _query(query, (code,), columns=True)
        return rset[0] if rset else rset
    elif by == 'sex':
        query = _load_sql_file(Q_EMPLOYMENT_SEX)
    elif by == 'age':
        query = _load_sql_file(Q_EMPLOYMENT_AGE)
    elif by == 'sector':
        query = _load_sql_file(Q_EMPLOYMENT_SECTOR)

    data = _query(query, (code,), columns=True)
    return _split_data(data)



# coding: utf-8

"""Pyris configuration

Retrieve user and password from the YAML configuration file for the database
access
"""

import os
import io

import yaml
from yaml import load as yload


_cfgfile = os.environ.get('PYRIS_APP_SETTINGS')

if _cfgfile is not None:
    with open(_cfgfile, 'r') as fobj:
        conf = yload(fobj.read(), yaml.Loader)
        DATABASE = conf.get('database', {})
        geocoder_conf = conf.get("geocoder")
        if geocoder_conf:
            GEOCODER_URL = geocoder_conf["url"]
        else:
            GEOCODER_URL = "http://api-adresse.data.gouv.fr"
else:
    DATABASE = {"USER": os.environ["USER"],
                "HOST": "localhost"}
    GEOCODER_URL = "http://api-adresse.data.gouv.fr"



# -*- coding: utf-8 -*-

import setuptools

with open("README.rst") as fobj:
    LONG_DESCRIPTION = fobj.read()

INSTALL_REQUIRES = ["Flask==1.0.2",
                    "flask-restplus==0.12.1",
                    "Werkzeug==0.16.1",
                    "psycopg2-binary",
                    "pyaml",
                    "slumber",
                    "gunicorn==19.9", ]

setuptools.setup(
    name='pyris',
    version='0.7',
    license='BSD',
    url='https://github.com/Oslandia/pyris',
    packages=setuptools.find_packages(exclude=['data', 'app.yml']),
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    # pip install -e .[dev]
    extras_require={'dev': ['pytest', 'pytest-sugar', 'ipython', 'ipdb']},

    author="Damien Garaud",
    author_email='damien.garaud@gmail.com',
    description="INSEE/IRIS geolocalization",
    long_description=LONG_DESCRIPTION,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ]
)



from flask_restplus import Namespace, Resource, fields

from pyris.api import extract


api = Namespace('insee', description='Some stats from INSEE for each IRIS')

population_parser = api.parser()
population_parser.add_argument("by", required=True, location='args',
                               help="By sex or age")
logement_parser = api.parser()
logement_parser.add_argument("by", required=True, location='args',
                             help="By room, area or year")
activite_parser = api.parser()
activite_parser.add_argument("by", required=True, location='args',
                             help="By sex, age")


@api.route('/')
class InseeData(Resource):
    @api.doc(description='INSEE data list')
    def get(self):
        return ['population', 'activite', 'logement',
                'menage', 'formation']


@api.route('/population/<string:code>')
class IrisPopulation(Resource):
    @api.doc("get the population for an IRIS")
    def get(self, code):
        if len(code) != 9:
            api.abort(400, "IRIS code is malformed (9 digits)")
        rset = extract.get_iris_population(code)
        if not rset:
            api.abort(404, "IRIS code '{}' not found".format(code))
        return rset[0]


@api.route('/population/distribution/<string:code>')
class IrisPopulationDistribution(Resource):
    @api.doc(parser=population_parser,
             description="get the population distribution for an IRIS")
    def get(self, code):
        if len(code) != 9:
            api.abort(400, "IRIS code is malformed (9 digits)")
        args = population_parser.parse_args()
        if args['by'] not in ('sex', 'age'):
            api.abort(400, "By query param should be 'sex' or 'age'")
        if args['by'] == 'sex':
            rset = extract.get_iris_population_sex(code)
        if args['by'] == 'age':
            rset = extract.get_iris_population_age(code)
        if not rset:
            api.abort(404, "IRIS code '{}' not found".format(code))
        return rset


@api.route('/logement/<string:code>')
class IrisLogement(Resource):
    @api.doc("get the housing data for an IRIS")
    def get(self, code):
        if len(code) != 9:
            api.abort(400, "IRIS code is malformed (9 digits)")
        rset = extract.get_iris_logement(code)
        if not rset:
            api.abort(404, "IRIS code '{}' not found".format(code))
        return rset


@api.route('/logement/distribution/<string:code>')
class IrisLogementDistribution(Resource):
    @api.doc(parser=logement_parser,
             description="get the housing data distribution for an IRIS")
    def get(self, code):
        if len(code) != 9:
            api.abort(400, "IRIS code is malformed (9 digits)")
        args = logement_parser.parse_args()
        if args['by'] not in ('room', 'area', 'year'):
            api.abort(400, "By query param should be 'year', 'room' or 'area'")
        rset = extract.get_iris_logement(code, by=args['by'])
        if not rset:
            api.abort(404, "IRIS code '{}' not found".format(code))
        return rset


@api.route('/activite/<string:code>')
class IrisActivite(Resource):
    @api.doc("get the employment data for an IRIS")
    def get(self, code):
        if len(code) != 9:
            api.abort(400, "IRIS code is malformed (9 digits)")
        rset = extract.get_iris_employment(code)
        if not rset:
            api.abort(404, "IRIS code '{}' not found".format(code))
        return rset


@api.route('/activite/distribution/<string:code>')
class IrisActiviteDistribution(Resource):
    @api.doc(parser=activite_parser,
             description="get the employment distribution data for an IRIS")
    def get(self, code):
        if len(code) != 9:
            api.abort(400, "IRIS code is malformed (9 digits)")
        args = activite_parser.parse_args()
        if args['by'] not in ('sex', 'age', 'sector'):
            api.abort(400, "By query param should be 'sex', 'age'")
        rset = extract.get_iris_employment(code, by=args['by'])
        if not rset:
            api.abort(404, "IRIS code '{}' not found".format(code))
        return rset
